import re

from markdown import Markdown, Extension
from markdown.preprocessors import Preprocessor


class InlineMathPreprocessor(Preprocessor):
    """
    Canvas supports \(...\) inline math (MathJax)
    But the \(...\) get mangled into (...) somewhere along the way.
    This extension stashes \(...\) so it is preserved
    """

    def run(self, lines):
        new_lines = []

        # inline_dollar_math_pattern = re.compile(r'(?<!\$)\$(?!\$)(.*?)(?<!\$)\$(?!\$)')
        # See https://docs.mathjax.org/en/latest/basic/mathematics.html
        #  for the rationale for why the $...$ syntax is not usually supported

        mathjax_inline_pattern = re.compile(r'\\\((.+?)\\\)')

        for line in lines:
            m = mathjax_inline_pattern.search(line)
            while m:
                line = line.replace(m.group(0), self.md.htmlStash.store(m.group(0)))
                m = mathjax_inline_pattern.search(line)
            new_lines.append(line)

        return new_lines


# Define the extension that uses the preprocessor
class InlineMathExtension(Extension):
    def extendMarkdown(self, md: Markdown):
        md.preprocessors.register(InlineMathPreprocessor(md), 'inlinemath', 27)



from canvasapi.course import Course
from canvasapi.discussion_topic import DiscussionTopic

from .util import get_canvas_object


def get_announcement(course: Course, title: str) -> DiscussionTopic:
    # NB: the `course` object here was modified in main.py to have a `canvas` field
    # That's why the following code works
    return get_canvas_object(
        lambda: course.canvas.get_announcements(context_codes=[f'course_{course.id}']),
        'title', title
    )


def deploy_announcement(course: Course, announcement_info: dict) -> tuple[DiscussionTopic, str | None]:
    title = announcement_info["title"]

    canvas_announcement: DiscussionTopic
    if canvas_announcement := get_announcement(course, title):
        canvas_announcement.update(**announcement_info)
    else:
        canvas_announcement = course.create_discussion_topic(**announcement_info)

    return canvas_announcement, None


def lookup_announcement(course: Course, announcement_name: str) -> DiscussionTopic:
    return get_announcement(course, announcement_name)



from canvasapi.course import Course
from canvasapi.assignment import AssignmentGroup
from ..our_logging import get_logger
logger = get_logger()


def get_group(course: Course, group_name: str) -> AssignmentGroup | None:
    for group in course.get_assignment_groups():
        if group.name == group_name:
            return group
    return None


def create_group(course: Course, group_name: str) -> AssignmentGroup:
    return course.create_assignment_group(name=group_name)


def update_group(group: AssignmentGroup, **kwargs) -> AssignmentGroup:
    return group.edit(**kwargs)


def deploy_group_weights(course: Course, group_weights: dict):
    # Update the course to apply the assignment group weights
    course.update(course={'apply_assignment_group_weights': True})

    # Set the group weights
    for group in group_weights:
        # Check if the group exists
        if (canvas_group := get_group(course, group)) is None:
            logger.info(f'Creating group: {group}')
            canvas_group = create_group(course, group)

        # Check if the groups weight needs to be updated
        if canvas_group.group_weight != group_weights[group]:
            logger.info(f'Updating group weight: {group}')
            update_group(canvas_group, group_weight=group_weights[group])




from canvasapi.assignment import Assignment
from canvasapi.course import Course

from .util import get_canvas_object, update_group_name_to_id


def get_assignment(course: Course, name: str) -> Assignment:
    return get_canvas_object(course.get_assignments, 'name', name)


def deploy_assignment(course: Course, assignment_info: dict) -> tuple[Assignment, str|None]:
    name = assignment_info["name"]

    update_group_name_to_id(course, assignment_info)

    # TODO - update group_category (name) to group_category_id
    #  Is this necessary to support?

    if canvas_assignment := get_assignment(course, name):
        canvas_assignment.edit(assignment=assignment_info)
    else:
        canvas_assignment = course.create_assignment(assignment=assignment_info)

    return canvas_assignment, None


def lookup_assignment(course: Course, assignment_name: str) -> Assignment:
    return get_assignment(course, assignment_name)



from bs4 import Tag

from .attributes import Attribute, parse_bool, parse_settings, parse_int
from ..resources import ResourceManager, get_key, CanvasResource


class ModuleTagProcessor:
    def __init__(self, resource_manager: ResourceManager):
        self._resources = resource_manager
        self._previous_module = None  # The name of the previous module

    _module_item_type_casing = {
        "file": "File",
        "page": "Page",
        "discussion": "Discussion",
        "assignment": "Assignment",
        "quiz": "Quiz",
        "subheader": "SubHeader",
        "externalurl": "ExternalUrl",
        "externaltool": "ExternalTool"
    }

    def __call__(self, module_tag: Tag):
        fields = [
            Attribute('title', required=True, new_name='name'),
            Attribute('position'),
            Attribute('published', parser=parse_bool)
        ]

        module_data = parse_settings(module_tag, fields)

        module_data['items'] = [
            self._parse_module_item(item_tag)
            for item_tag in module_tag.find_all('item')
        ]

        if self._previous_module is not None:
            # adding a reference to the previous module ensures this module
            #  is created after the previous one, thus preserving their
            #  relative ordering
            module_data['_comments'] = {
                'previous_module': get_key('module', self._previous_module, 'id')
            }

        self._previous_module = module_data['name']

        self._resources.add_resource(CanvasResource(
            type='module',
            name=module_data['name'],
            data=module_data
        ))

    def _parse_module_item(self, tag: Tag) -> dict:
        fields = [
            Attribute('type', ignore=True),
            Attribute('title', required=True),
            Attribute('position', parser=parse_int),
            Attribute('indent', parser=parse_int),
            Attribute('new_tab', True, parse_bool),
            Attribute('completion_requirement'),
            Attribute('iframe'),
            Attribute('published', parser=parse_bool),
        ]

        name = tag['title']
        rtype = self._module_item_type_casing[tag['type'].lower()]

        item = {
            'type': rtype
        }

        if rtype == 'Page':
            item['page_url'] = get_key(rtype.lower(), name, 'url')

        elif rtype == 'ExternalUrl':
            fields.append(Attribute(
                'external_url', required=True
            ))

        elif rtype == 'SubHeader':
            pass  # TODO - fix the fields for this (if necessary?)

        else:
            item['id'] = get_key(rtype.lower(), name, 'id')

        item.update(parse_settings(tag, fields))

        return item



import re
import string

from bs4 import Tag

from .attributes import parse_settings, Attribute, parse_bool, parse_int, parse_children_tag_contents
from ..util import retrieve_contents

NO_POINTS = 0
FULL_POINTS = 100

question_children_names = [
    'correct', 'incorrect', 'pair', 'distractors', 'correct-comments', 'neutral-comments', 'incorrect-comments'
]


common_fields = [
    Attribute('correct-comments', new_name='correct_comments'),
    Attribute('neutral-comments', new_name='neutral_comments'),
    Attribute('incorrect-comments', new_name='incorrect_comments'),
    Attribute('text-after-answers', new_name='text_after_answers'),
    Attribute('type', ignore=True),
    Attribute('name', ignore=True)
]

mostly_common_fields = [
    Attribute('points', 1, parse_int, 'points_possible'),
    *common_fields
]


def parse_text_question(tag: Tag):
    question_text = retrieve_contents(tag)
    question = {
        "question_text": question_text,
        "question_type": 'text_only_question',
    }
    return [question]


def parse_true_false_question(tag: Tag):
    """
    <question type='true-false' answer='true' points_possible='2'>
    The earth orbits the sun
    </question>

    <question type='true-false' answer='false'>
    The earth is **flat**

    <correct-comments>
    A nationwide survey in 2022 by researchers at the University of New Hampshire found that
    10% of U.S. adults believed the earth was flat.
    </correct-comments>

    <incorrect-comments>
    Regular folks who like math and stars rely on the curvature on the earth to track the motion
    of heavenly bodies.
    </incorrect-comments>
    </question>
    """
    fields = [
        Attribute('answer', required=True, parser=parse_bool, default=False)
    ]
    question = parse_settings(tag, mostly_common_fields + fields)

    question.update({
        "question_text": retrieve_contents(tag, question_children_names),
        "question_type": 'true_false_question',
        "answers": [
            {
                "answer_text": "True",
                "answer_weight": FULL_POINTS if question["answer"] is True else NO_POINTS
            },
            {
                "answer_text": "False",
                "answer_weight": FULL_POINTS if question["answer"] is False else NO_POINTS
            }
        ]
    })

    return [question]


def parse_multiple_choice_question(tag: Tag):
    """
    <question type='multiple-choice'>
    5 + 5 =
    <correct> 10 </correct>
    <incorrect> 11 </incorrect>
    <incorrect> 9 </incorrect>
    <incorrect> 8 </incorrect>
    </question>
    """
    return _parse_multiple_option_question('multiple_choice_question', tag)


def parse_multiple_answers_question(tag: Tag):
    """
    <question type='multiple-answers'>
    Which of the following are prime numbers?
    <correct> 2 </correct>
    <correct> 3 </correct>
    <incorrect> 4 </incorrect>
    <correct> 5 </correct>
    <incorrect> 6 </incorrect>
    </question>
    """
    return _parse_multiple_option_question('multiple_answers_question', tag)


def _parse_multiple_option_question(question_type, tag):
    corrects = parse_children_tag_contents(tag, 'correct')
    answers = parse_children_tag_contents(tag, re.compile(r'correct|incorrect'))
    question = {
        "question_text": retrieve_contents(tag, question_children_names),
        "question_type": question_type,
        "answers": [
            {
                "answer_html": answer,
                "answer_weight": FULL_POINTS if answer in corrects else NO_POINTS
            } for answer in answers
        ]
    }
    question.update(parse_settings(tag, mostly_common_fields))
    return [question]


def parse_matching_question(tag: Tag):
    """
    <question type='matching'>
    Match the following:
    <pair left='1' right='A' />
    <pair left='2' right='B' />
    <pair left='3' right='C' /
    <distractors>
    D
    E
    </distractors>
    <correct-comments>Good job!</correct-comments>
    </question>
    """
    left_field = Attribute('left', required=True)
    right_field = Attribute('right', required=True)
    pairs = [
        parse_settings(answer, [left_field, right_field]) for answer in tag.find_all('pair')
    ]
    distractors = '\n'.join(parse_children_tag_contents(tag, 'distractors'))
    distractors = '\n'.join(line for line in distractors.splitlines() if line.split())

    question = parse_settings(tag, mostly_common_fields)

    question.update({
        "question_text": retrieve_contents(tag, question_children_names + ['pair', 'distractors']),
        "question_type": 'matching_question',
        "points_possible": parse_int(tag.get('points') or len(pairs)),
        "answers": [
            {
                "answer_match_left": answer['left'],
                "answer_match_right": answer['right'],
                "answer_weight": FULL_POINTS
            } for answer in pairs
        ],
        "matching_answer_incorrect_matches": distractors
    })

    return [question]


def parse_multiple_true_false_question(tag: Tag):
    """
    <question type='multiple-tf'>
    Which of the following matrices are invertible?

    A: [[1, 0], [0, 1]]
    B: [[1, 0], [1, 0]]
    C: [[1, 1], [1, 1]]
    D: [[0, 0], [0, 0]]

    <correct> A </correct>
    <incorrect> B </incorrect>
    <incorrect> C </incorrect>
    <incorrect> D </incorrect>
    """
    resulting_questions = []

    header = retrieve_contents(tag, question_children_names)
    resulting_questions.append({
        "question_text": header,
        "question_type": 'text_only_question',
    })

    answers = tag.find_all(re.compile(r'correct|incorrect'))

    question_attributes = [
        Attribute('points', default=len(answers), parser=parse_int, new_name='points_possible'),
    ]
    settings = parse_settings(tag, question_attributes + common_fields)
    total_points = settings['points_possible']

    # Distribute points evenly among the questions
    points_per_question = round(total_points // len(answers), 2)
    error = total_points - points_per_question * len(answers)
    num_to_change = round(error * 100)
    amount = 0.01 if error > 0 else -0.01

    for index, child in enumerate(answers):
        resulting_questions.append({
            "question_text": retrieve_contents(child),
            "question_type": 'true_false_question',
            "points_possible": points_per_question + amount if index < num_to_change else points_per_question,
            "answers": [
                {
                    "answer_text": "True",
                    "answer_weight": FULL_POINTS if child.name == 'correct' else NO_POINTS
                },
                {
                    "answer_text": "False",
                    "answer_weight": FULL_POINTS if child.name == 'incorrect' else NO_POINTS
                }
            ]
        })

    return resulting_questions

def _add_answers_to_multiple_blanks_question(text):
    def letter_generator():
        for repeat in range(1, 10):
            for letter in string.ascii_uppercase:
                yield letter * repeat
    letter_generator = letter_generator()
    answers = []
    def get_next_letter(match):
        answer = match.group()[1:-1]
        associated_id = next(letter_generator)
        answers.append({'answer_text': answer, 'blank_id': associated_id, 'answer_weight': FULL_POINTS})
        return f'[{associated_id}]'

    updated_text = re.sub(r"\[(.*?)\]", get_next_letter, text)

    return updated_text, answers

def parse_fill_in_multiple_blanks_filled_answers(tag: Tag):
    """
    Anything within a set of brackets will be turned into a fill in the blank question.
    Whatever is in the brackets will be the correct answer

    The default number of points is the number of fill in the blank questions, but can be overwritten using "points".

    <question type='fill-in-multiple-blanks-filled-answers' points="5">
            The U.S. flag has [13] stripes and [50] stars.
    </question>

    This is also useful for tables
    <question type="fill-in-multiple-blanks-filled-answers">
        Fill in the table using the algorithm discussed in class.

        | Node | 0     | 1     | 2     |
        |------|-------|-------|-------|
        | A    | 0     | [0]   | [0]   |
        | B    | [inf] | [1]   | [1]   |
        | C    | [inf] | [inf] | [3]   |
        | D    | [inf] | inf   | [inf] |
        | E    | [inf] | [4]   | 4     |
        | F    | [inf] | 8     | [7]   |
        | G    | inf   | [inf] | [7]   |
        | H    | [inf] | [inf] | [inf] |

    </question>
    """
    question_text = retrieve_contents(tag, question_children_names)

    question_text, answers = _add_answers_to_multiple_blanks_question(question_text)

    question = {
        "question_text": question_text,
        "question_type": 'fill_in_multiple_blanks_question',
        "answers": answers,
    }

    question_attributes = [
        Attribute('points', default=len(answers), parser=parse_int, new_name='points_possible'),
    ]
    settings = parse_settings(tag, question_attributes + common_fields)
    question.update(settings)
    return [question]

def parse_fill_in_the_blank_question(tag: Tag):
    """
    <question type='fill-in-the-blank'>
    The capital of France is [blank].
    <correct text='Paris' />
    </question>
    """
    question_text = retrieve_contents(tag, question_children_names)
    blanks = re.findall(r'\[([^]]+)]', question_text)
    if not blanks:
        raise ValueError("Fill in the blank questions must contain at least one blank!")

    answer_attributes = [
        Attribute('text', required=True, new_name='answer_text'),
        Attribute('blank_id', blanks[0]),
        Attribute('answer_weight', FULL_POINTS),
    ]

    question = {
        "question_text": retrieve_contents(tag, question_children_names),
        "question_type": 'fill_in_multiple_blanks_question',
        "answers": [
            parse_settings(answer, answer_attributes) for answer in tag.find_all('correct')
        ]
    }

    question.update(parse_settings(tag, mostly_common_fields))
    return [question]


def parse_fill_in_multiple_blanks_question(tag: Tag):
    """
    <question type='fill-in-multiple-blanks'>
    The U.S. flag has [stripes] stripes and [stars] stars.
    <correct text='13' blank='stripes' />
    <correct text='50' blank='stars' />
    </question>

    The default number of points is the number of fill in the blank questions (ie the example above would be worth 2 points)
    """
    answer_attributes = [
        Attribute('text', required=True, new_name='answer_text'),
        Attribute('blank', required=True, new_name='blank_id'),
        Attribute('answer_weight', FULL_POINTS)
    ]

    answers = tag.find_all('correct')
    question = {
        "question_text": retrieve_contents(tag, question_children_names),
        "question_type": 'fill_in_multiple_blanks_question',
        "answers": [
            parse_settings(answer, answer_attributes) for answer in answers
        ]
    }

    question_attributes = [
        Attribute('points', default=len(answers), parser=parse_int, new_name='points_possible'),
    ]
    settings = parse_settings(tag, question_attributes + common_fields)
    question.update(settings)
    return [question]


def parse_essay_question(tag: Tag):
    question_text = retrieve_contents(tag)
    question = {
        "question_text": question_text,
        "question_type": 'essay_question',
    }
    question.update(parse_settings(tag, mostly_common_fields))
    return [question]


def parse_file_upload_question(tag: Tag):
    question_text = retrieve_contents(tag)
    question = {
        "question_text": question_text,
        "question_type": 'file_upload_question',
    }
    return [question]


def parse_exact_answer_question(tag: Tag):
    """
    <question type='numerical' numerical_answer_type="exact">
    Give one possible value for x. The margin of error is +- 0.0001.

    (x - pi)^2 = (x - pi)

    <correct answer_exact='3.14159' answer_error_margin='0.0001' />

    <correct answer_exact='4.14159' answer_error_margin='0.0001' />
    </question>
    """

    question_text = retrieve_contents(tag, ['answer_exact', 'answer_error_margin'])
    answer_attributes = [
        Attribute('answer_exact', required=True),
        Attribute('answer_error_margin', required=True)
    ]

    return question_text, answer_attributes


def parse_range_answer_question(tag: Tag):
    """
    <question type='numerical' numerical_answer_type="range">
    Give one possible value for x.

    1 <= x^2 <= 100

    <correct answer_range_start='1' answer_range_end='10' />

    <correct answer_range_start='-10' answer_range_end='-1' />
    </question>
    """
    question_text = retrieve_contents(tag, ['answer_range_start', 'answer_range_end'])

    answer_attributes = [
        Attribute('answer_range_start', required=True),
        Attribute('answer_range_end', required=True)
    ]

    return question_text, answer_attributes


def parse_precision_answer_question(tag: Tag):
    """
    The precision number is how many digits are expected in the answer.
    Precision answers can be negative numbers and may include trailing zeroes.
    However, student responses will be marked as correct if they omit the trailing zeroes, as long as all preceding digits are correct.

    <question type='numerical' numerical_answer_type="precision">
    What is the value of pi?

    Ensure your answer gives at least 5 digits.

    <correct answer_approximate='3.14159' answer_precision='5' />
    </question>
    """

    question_text = retrieve_contents(tag, ['answer_approximate', 'answer_precision'])
    answer_attributes = [
        Attribute('answer_approximate', required=True),
        Attribute('answer_precision', required=True)
    ]

    return question_text, answer_attributes


def parse_numerical_question(tag: Tag):
    numerical_answer_types = {
        'exact': (parse_exact_answer_question, 'exact_answer'),
        'range': (parse_range_answer_question, 'range_answer'),
        'precision': (parse_precision_answer_question, 'precision_answer')
    }

    numerical_answer_type = tag.get('numerical_answer_type')
    if numerical_answer_type not in numerical_answer_types:
        raise ValueError(f"Invalid numerical answer type: {numerical_answer_type}")

    parse_function, answer_type = numerical_answer_types[numerical_answer_type]
    question_text, answer_attributes = parse_function(tag)

    question = {
        "question_text": question_text,
        "question_type": 'numerical_question',
        "answers": [
            parse_settings(answer, answer_attributes) for answer in tag.find_all('correct')
        ]
    }

    fields = [
        Attribute('numerical_answer_type', required=True, ignore=True)
    ]
    question.update(parse_settings(tag, fields + common_fields))

    for answer in question["answers"]:
        answer["numerical_answer_type"] = answer_type

    question.update(parse_settings(tag, mostly_common_fields))
    return [question]




from collections import deque

from ..our_logging import get_logger

logger = get_logger()


def linearize_dependencies(graph: dict[tuple[str, str], list[tuple[str, str]]]) -> list[tuple[str, str]]:
    # Original code courtesy of GPT-4o
    # Some modifications required

    in_degree = {}

    # Build the graph and compute in-degrees of nodes
    missing_deps = []
    for key, deps in graph.items():
        if key not in in_degree:
            in_degree[key] = 0
        for dep in deps:
            if dep not in in_degree:
                in_degree[dep] = 0
            if dep not in graph:
                missing_deps.append(dep)
            in_degree[dep] += 1

    # Perform topological sort (Kahn's algorithm)
    queue = deque([key for key in in_degree if in_degree[key] == 0])
    linearized_order = []

    while queue:
        current_key = queue.popleft()
        linearized_order.append(current_key)

        for dependent_key in graph.get(current_key, []):
            in_degree[dependent_key] -= 1
            if in_degree[dependent_key] == 0:
                queue.append(dependent_key)

    # Check for cycles (if not all keys are in the linearized order)
    if any(v != 0 for v in in_degree.values()):
        logger.error(graph)
        logger.error(in_degree)
        raise ValueError("Dependency graph has at least one cycle")

    return linearized_order[::-1]


if __name__ == '__main__':
    dependency_dict = {
        'A': ['B'],
        'B': ['C', 'D'],
        'C': [],
        'D': ['C']
    }

    # expect C, D, B, A
    order = linearize_dependencies(dependency_dict)
    print(order)



import dataclasses
from datetime import datetime
from typing import Callable, Any

from bs4 import Tag

from ..our_logging import get_logger
from ..util import retrieve_contents

logger = get_logger()


def parse_date(date: datetime | str | None) -> str:
    if isinstance(date, datetime):
        return datetime.isoformat(date)

    elif isinstance(date, str):
        # Check if the string is already in ISO format
        try:
            return datetime.isoformat(datetime.fromisoformat(date))
        except ValueError:
            pass

        try_formats = [
            "%b %d, %Y, %I:%M %p",
            "%b %d %Y %I:%M %p",
            "%Y-%m-%dT%H:%M:%S%z"
        ]
        for format_str in try_formats:
            try:
                parsed_date = datetime.strptime(date, format_str)
                break
            except ValueError:
                pass
        else:
            raise ValueError(f"Invalid date format: {date}")

        # Convert the parsed datetime object to the desired timezone
        # to_zone = pytz.timezone(time_zone)
        # parsed_date = parsed_date.replace(tzinfo=None)  # Remove existing timezone info
        # parsed_date = parsed_date.astimezone(to_zone)
        # TODO - Move this timezone block to deployment
        return datetime.isoformat(parsed_date)
    else:
        raise TypeError("Date must be a datetime object or a string")


def parse_int(text):
    return int(text)


def parse_bool(text):
    # Forgiving boolean parser
    if isinstance(text, bool):
        return text

    if text.lower() == "true":
        return True
    elif text.lower() == "false":
        return False
    else:
        raise ValueError(f"Invalid boolean value: {text}")


def parse_list(text):
    items = text.strip().split(',')
    return [cell.strip() for cell in items if cell.strip()]


def parse_dict(text):
    # Assumes the string is a comma-separated list of key-value pairs
    # Example: "key1=value1, key2=value2 "
    return dict(cell.strip().split('=') for cell in text.split(',') if cell.strip())


@dataclasses.dataclass
class Attribute:
    name: str
    default: Any = None
    parser: Callable[[str], Any] = lambda x: x
    new_name: str = None
    required: bool = False
    ignore: bool = False
    is_tag: bool = False


def get_tag_info(tag: Tag):
    name = tag.name
    hint = tag.get('title', None)
    if hint is None:
        hint = tag.get('name', None)
    display = name
    if hint is not None:
        display += f'({hint})'
    return display


def get_tag_path(tag: Tag):
    tokens = [get_tag_info(tag)] + [get_tag_info(p) for p in tag.parents]
    return '.'.join(tokens[::-1])


def parse_settings(tag: Tag, attributes: list[Attribute]):
    settings = {}
    processed_fields = set()

    for attribute in attributes:
        try:

            processed_fields.add(attribute.name)
            if attribute.new_name:
                processed_fields.add(attribute.new_name)
            name = attribute.new_name or attribute.name

            if attribute.ignore:
                continue

            if (field := (
                    tag.get(attribute.name, None)
                    or tag.get(attribute.new_name, None)
            )) is not None:
                # Parse the value
                value = attribute.parser(field)
                settings[name] = value

            elif attribute.is_tag:
                child = tag.find(attribute.name, recursive=False)
                value = retrieve_contents(child)
                settings[name] = attribute.parser(value)

            elif attribute.default is not None:
                settings[name] = attribute.default

            elif attribute.required:
                raise Exception(f'Required field {attribute.name} missing from tag {tag.name} @ {get_tag_path(tag)}')
        except:
            logger.exception(f"Error processing attribute {attribute.name} in tag {get_tag_path(tag)}")
            raise

    for key in tag.attrs:
        if key not in processed_fields:
            logger.warning(f'Unprocessed_fields field {key} @ {get_tag_path(tag)}')

    return settings


def parse_children_tag_contents(tag: Tag, child_name):
    children = tag.find_all(child_name, recursive=False)
    return [retrieve_contents(child) for child in children]



from pathlib import Path
from typing import Callable

from bs4 import Tag

from ..resources import ResourceManager, FileData, ZipFileData, CanvasResource, get_key
from ..util import parse_soup_from_xml
from ..xml_processing.attributes import parse_bool, get_tag_path


def make_image_preprocessor(parent: Path, resources: ResourceManager):
    def process_image(tag: Tag):
        # TODO - handle b64-encoded images

        src = tag.get('src')
        if src.startswith('http'):
            # No changes necessary
            return

        # Assume it's a local file
        src = (parent / src).resolve().absolute()
        if not src.is_file():
            raise ValueError(f"Image file {src} is not a file @ {get_tag_path(tag)}")

        file = CanvasResource(
            type='file',
            name=src.name,
            data=FileData(
                path=str(src),
                canvas_folder=tag.get('canvas_folder', None)
            )
        )
        tag['src'] = resources.add_resource(file, 'uri') + '/preview'

    return process_image


def make_file_anchor_tag(resource_key: str, filename: str, **kwargs):
    attrs = {
        **kwargs,
        'href': f'{resource_key}?wrap=1',
        'class': 'instructure_file_link inline_disabled',
        'title': filename,
        'target': '_blank',
        'rel': 'noopener noreferrer'
    }

    new_tag = Tag(name='a', attrs=attrs)
    new_tag.string = filename

    return new_tag


def make_file_preprocessor(parent: Path, resources: ResourceManager):
    def process_file(tag: Tag):
        attrs = tag.attrs
        path = (parent / attrs.pop('path')).resolve().absolute()
        if not path.is_file():
            raise ValueError(f"File {path} is not a file @ {get_tag_path(tag)}")
        file = CanvasResource(
            type='file',
            name=path.name,
            data=FileData(
                path=str(path),
                canvas_folder=attrs.get('canvas_folder', None),
                unlock_at=attrs.get('unlock_at', None),
                lock_at=attrs.get('lock_at', None)
            )
        )
        resource_key = resources.add_resource(file, 'uri')
        new_tag = make_file_anchor_tag(resource_key, path.name, **tag.attrs)

        tag.replace_with(new_tag)

    return process_file


def make_zip_preprocessor(parent: Path, resources: ResourceManager):
    def process_zip(tag: Tag):
        content_folder = tag.get("path")

        name = tag.get("name")
        if not name:
            name = (
                       content_folder
                       .replace('.', '')
                       .replace('/', '-')
                       .strip('-')
                   ) + '.zip'

        content_folder = str((parent / content_folder).resolve().absolute())

        additional_files = tag.get("additional_files")
        if additional_files:
            additional_files = [str((parent / file).resolve().absolute()) for file in additional_files.split(',')]

        priority_folder = tag.get("priority_path")
        if priority_folder:
            priority_folder = str((parent / priority_folder).resolve().absolute())

        exclude_pattern = tag.get("exclude")

        file = CanvasResource(
            type='zip',
            name=name,
            data=ZipFileData(
                zip_file_name=name,
                content_folder=content_folder,
                additional_files=additional_files,
                exclude_pattern=exclude_pattern,
                priority_folder=priority_folder,
                canvas_folder=tag.get('canvas_folder')
            )
        )

        resource_key = resources.add_resource(file, 'uri')

        new_tag = make_file_anchor_tag(resource_key, name)
        tag.replace_with(new_tag)

    return process_zip


def _parse_slice(field: str) -> slice:
    """
    Parse a 1-based, inclusive slice
    So, the slice should match the line numbers shown in your IDE
    """
    tokens = field.split(':')
    tokens = [
        int(token) if token else None
        for token in tokens
    ]

    tokens[0] -= 1  # make it 1-based

    if len(tokens) == 1:  # e.g. "3"
        tokens.append(None)

    # Tokens[1] +1 for inclusive, -1 for one-based, net: 0

    return slice(tokens[0], tokens[1])


def make_include_preprocessor(
        parent_folder: Path,
        process_file: Callable
):
    def process_include(tag: Tag):
        imported_filename = tag.get('path')
        imported_file = (parent_folder / imported_filename).resolve()

        imported_raw_content = imported_file.read_text(encoding='utf-8')
        suffixes = imported_file.suffixes

        lines = tag.get('lines', '')
        if lines:
            grab = _parse_slice(lines)
            imported_raw_content = '\n'.join(imported_raw_content.splitlines()[grab])

        if parse_bool(tag.get('fenced', 'false')):
            imported_raw_content = f'```{imported_file.suffix.lstrip(".")}\n{imported_raw_content}\n```\n'
            suffixes = suffixes + ['.md']

        args_file = tag.get('args', None)
        if args_file is not None:
            # Note: the args file, like the path, is relative to the primary file, not the included file
            args_file = (parent_folder / args_file).resolve().absolute()

        imported_html = process_file(
            imported_file.parent,
            imported_raw_content,
            suffixes,
            args_file=args_file
        )

        use_div = parse_bool(tag.get('usediv', 'true'))

        include_result = parse_soup_from_xml(imported_html)

        if not use_div:
            tag.parent.extend(include_result)

        else:
            new_tag = Tag(name='div')
            new_tag['data-source'] = imported_filename
            if lines:
                new_tag['data-lines'] = lines
            new_tag.extend(include_result)
            tag.replace_with(new_tag)

    return process_include


def make_link_preprocessor():
    def process_link(tag: Tag):
        link_type = tag['type']
        link_title = tag['title']

        new_tag = Tag(name='a')
        new_tag['href'] = get_key(link_type, link_title, 'uri')
        # TODO - add other course-link attributes here
        link_text = tag.string.strip() if tag.string is not None else ''
        if not link_text:
            link_text = tag['title']
        new_tag.string = link_text
        tag.replace_with(new_tag)

    return process_link


def make_markdown_page_preprocessor(
        parent_folder: Path,
        process_file: Callable
):
    def process_markdown_page(tag: Tag):
        content_path = tag['path']

        page_title = tag.get('title')
        if page_title is None:
            page_title = Path(content_path).stem

        include_tag = Tag(name='include', attrs={'path': content_path})

        page_tag = Tag(name='page', attrs={'title': page_title})
        page_tag.append(include_tag)

        include_processor = make_include_preprocessor(parent_folder, process_file)
        include_processor(include_tag)  # Replaces include_tag with new content

        tag.replace_with(page_tag)

    return process_markdown_page



from canvasapi.course import Course
from canvasapi.page import Page

from .util import get_canvas_object, get_canvas_uri


def get_page(course: Course, title: str) -> Page:
    return get_canvas_object(course.get_pages, 'title', title)


def deploy_page(course: Course, page_info: dict) -> tuple[Page, str|None]:
    name = page_info['title']

    if canvas_page := get_page(course, name):
        canvas_page.edit(wiki_page=page_info)
    else:
        canvas_page = course.create_page(wiki_page=page_info)

    return canvas_page, None


def lookup_page(course: Course, page_title: str) -> Page:
    canvas_page = get_page(course, page_title)
    if not canvas_page:
        raise Exception(f'Page {page_title} not found')

    return canvas_page



from canvasapi.assignment import Assignment, AssignmentOverride
from canvasapi.course import Course

from .assignment import get_assignment
from .util import get_canvas_object


def get_override(assignment: Assignment, section_id: int) -> AssignmentOverride:
    return get_canvas_object(assignment.get_overrides, 'course_section_id', section_id)


def get_assignment_and_section(course: Course, name: str) -> tuple[Assignment, int]:
    assignment_name, section_id = name.split('|')
    assignment = get_assignment(course, assignment_name)
    return assignment, int(section_id)


def deploy_override(course: Course, override_info: dict) -> tuple[AssignmentOverride, str | None]:
    name = override_info["name"]

    # Convert assignment name to assignment_id
    assignment, section_id = get_assignment_and_section(course, name)

    if override := get_override(assignment, section_id):
        override.edit(assignment_override=override_info)
    else:
        override = assignment.create_override(assignment_override=override_info)

    return override, None


def lookup_override(course: Course, override_name: str) -> AssignmentOverride:
    assignment, section_id = get_assignment_and_section(course, override_name)
    return get_override(assignment, section_id)
